#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/epoll.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <vector>
#include <string>
#include <map>
#include <algorithm>

#include "markpool.h"
#include "metacube.h"
#include "server.h"
#include "mutexlock.h"
#include "parse.h"
#include "state.pb.h"

using namespace std;

Client::Client(int sock)
	: sock(sock),
	  fwmark(0),
	  connect_time(time(NULL)),
	  state(Client::READING_REQUEST),
	  stream(NULL),
	  header_or_error_bytes_sent(0),
	  bytes_sent(0)
{
	request.reserve(1024);

	// Find the remote address, and convert it to ASCII.
	sockaddr_in6 addr;
	socklen_t addr_len = sizeof(addr);

	if (getpeername(sock, reinterpret_cast<sockaddr *>(&addr), &addr_len) == -1) {
		perror("getpeername");
		remote_addr = "";
	} else {
		char buf[INET6_ADDRSTRLEN];
		if (inet_ntop(addr.sin6_family, &addr.sin6_addr, buf, sizeof(buf)) == NULL) {
			perror("inet_ntop");
			remote_addr = "";
		} else {
			remote_addr = buf;
		}
	}
}
	
Client::Client(const ClientProto &serialized, Stream *stream)
	: sock(serialized.sock()),
	  remote_addr(serialized.remote_addr()),
	  connect_time(serialized.connect_time()),
	  state(State(serialized.state())),
	  request(serialized.request()),
	  stream_id(serialized.stream_id()),
	  stream(stream),
	  header_or_error(serialized.header_or_error()),
	  header_or_error_bytes_sent(serialized.header_or_error_bytes_sent()),
	  bytes_sent(serialized.bytes_sent())
{
	if (stream->mark_pool != NULL) {
		fwmark = stream->mark_pool->get_mark();
	} else {
		fwmark = 0;  // No mark.
	}
	if (setsockopt(sock, SOL_SOCKET, SO_MARK, &fwmark, sizeof(fwmark)) == -1) {
		if (fwmark != 0) {
			perror("setsockopt(SO_MARK)");
		}
	}
}

ClientProto Client::serialize() const
{
	ClientProto serialized;
	serialized.set_sock(sock);
	serialized.set_remote_addr(remote_addr);
	serialized.set_connect_time(connect_time);
	serialized.set_state(state);
	serialized.set_request(request);
	serialized.set_stream_id(stream_id);
	serialized.set_header_or_error(header_or_error);
	serialized.set_header_or_error_bytes_sent(serialized.header_or_error_bytes_sent());
	serialized.set_bytes_sent(bytes_sent);
	return serialized;
}
	
ClientStats Client::get_stats() const
{
	ClientStats stats;
	stats.stream_id = stream_id;
	stats.remote_addr = remote_addr;
	stats.connect_time = connect_time;
	stats.bytes_sent = bytes_sent;
	return stats;
}

Stream::Stream(const string &stream_id)
	: stream_id(stream_id),
	  data(new char[BACKLOG_SIZE]),
	  data_size(0),
	  mark_pool(NULL)
{
	memset(data, 0, BACKLOG_SIZE);
}

Stream::~Stream()
{
	delete[] data;
}

Stream::Stream(const StreamProto &serialized)
	: stream_id(serialized.stream_id()),
	  header(serialized.header()),
	  data(new char[BACKLOG_SIZE]),
	  data_size(serialized.data_size()),
	  mark_pool(NULL)
{
	assert(serialized.data().size() == BACKLOG_SIZE);
	memcpy(data, serialized.data().data(), BACKLOG_SIZE);
}

StreamProto Stream::serialize() const
{
	StreamProto serialized;
	serialized.set_header(header);
	serialized.set_data(string(data, data + BACKLOG_SIZE));
	serialized.set_data_size(data_size);
	serialized.set_stream_id(stream_id);
	return serialized;
}

void Stream::put_client_to_sleep(Client *client)
{
	sleeping_clients.push_back(client);
}

void Stream::wake_up_all_clients()
{
	if (to_process.empty()) {
		swap(sleeping_clients, to_process);
	} else {
		to_process.insert(to_process.end(), sleeping_clients.begin(), sleeping_clients.end());
		sleeping_clients.clear();
	}
}

Server::Server()
{
	pthread_mutex_init(&mutex, NULL);
	pthread_mutex_init(&queued_data_mutex, NULL);

	epoll_fd = epoll_create(1024);  // Size argument is ignored.
	if (epoll_fd == -1) {
		perror("epoll_fd");
		exit(1);
	}
}

Server::~Server()
{
	int ret;
	do {
		ret = close(epoll_fd);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		perror("close(epoll_fd)");
	}
}

vector<ClientStats> Server::get_client_stats() const
{
	vector<ClientStats> ret;

	MutexLock lock(&mutex);
	for (map<int, Client>::const_iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		ret.push_back(client_it->second.get_stats());
	}
	return ret;
}

void Server::do_work()
{
	for ( ;; ) {
		int nfds = epoll_wait(epoll_fd, events, EPOLL_MAX_EVENTS, EPOLL_TIMEOUT_MS);
		if (nfds == -1 && errno == EINTR) {
			if (should_stop) {
				return;
			}
			continue;
		}
		if (nfds == -1) {
			perror("epoll_wait");
			exit(1);
		}

		MutexLock lock(&mutex);  // We release the mutex between iterations.
	
		process_queued_data();

		for (int i = 0; i < nfds; ++i) {
			int fd = events[i].data.fd;
			assert(clients.count(fd) != 0);
			Client *client = &clients[fd];

			if (events[i].events & (EPOLLERR | EPOLLRDHUP | EPOLLHUP)) {
				close_client(client);
				continue;
			}

			process_client(client);
		}

		for (map<string, Stream *>::iterator stream_it = streams.begin();
		     stream_it != streams.end();
		     ++stream_it) {
			vector<Client *> to_process;
			swap(stream_it->second->to_process, to_process);
			for (size_t i = 0; i < to_process.size(); ++i) {
				process_client(to_process[i]);
			}
		}

		if (should_stop) {
			return;
		}
	}
}

CubemapStateProto Server::serialize()
{
	// We don't serialize anything queued, so empty the queues.
	process_queued_data();

	CubemapStateProto serialized;
	for (map<int, Client>::const_iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		serialized.add_clients()->MergeFrom(client_it->second.serialize());
	}
	for (map<string, Stream *>::const_iterator stream_it = streams.begin();
	     stream_it != streams.end();
	     ++stream_it) {
		serialized.add_streams()->MergeFrom(stream_it->second->serialize());
	}
	return serialized;
}

void Server::add_client_deferred(int sock)
{
	MutexLock lock(&queued_data_mutex);
	queued_add_clients.push_back(sock);
}

void Server::add_client(int sock)
{
	clients.insert(make_pair(sock, Client(sock)));

	// Start listening on data from this socket.
	epoll_event ev;
	ev.events = EPOLLIN | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = sock;
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}

	process_client(&clients[sock]);
}

void Server::add_client_from_serialized(const ClientProto &client)
{
	MutexLock lock(&mutex);
	Stream *stream = find_stream(client.stream_id());
	clients.insert(make_pair(client.sock(), Client(client, stream)));
	Client *client_ptr = &clients[client.sock()];

	// Start listening on data from this socket.
	epoll_event ev;
	if (client.state() == Client::READING_REQUEST) {
		ev.events = EPOLLIN | EPOLLET | EPOLLRDHUP;
	} else {
		// If we don't have more data for this client, we'll be putting it into
		// the sleeping array again soon.
		ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	}
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client.sock();
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, client.sock(), &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}

	if (client_ptr->state == Client::SENDING_DATA && 
	    client_ptr->bytes_sent == client_ptr->stream->data_size) {
		client_ptr->stream->put_client_to_sleep(client_ptr);
	} else {
		process_client(client_ptr);
	}
}

void Server::add_stream(const string &stream_id)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream_id, new Stream(stream_id)));
}

void Server::add_stream_from_serialized(const StreamProto &stream)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream.stream_id(), new Stream(stream)));
}
	
void Server::set_header(const string &stream_id, const string &header)
{
	MutexLock lock(&mutex);
	find_stream(stream_id)->header = header;

	// If there are clients we haven't sent anything to yet, we should give
	// them the header, so push back into the SENDING_HEADER state.
	for (map<int, Client>::iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		Client *client = &client_it->second;
		if (client->state == Client::SENDING_DATA &&
		    client->bytes_sent == 0) {
			construct_header(client);
		}
	}
}
	
void Server::set_mark_pool(const std::string &stream_id, MarkPool *mark_pool)
{
	MutexLock lock(&mutex);
	assert(clients.empty());
	find_stream(stream_id)->mark_pool = mark_pool;
}

void Server::add_data_deferred(const string &stream_id, const char *data, size_t bytes)
{
	MutexLock lock(&queued_data_mutex);
	queued_data[stream_id].append(string(data, data + bytes));
}

void Server::add_data(const string &stream_id, const char *data, size_t bytes)
{
	Stream *stream = find_stream(stream_id);
	size_t pos = stream->data_size % BACKLOG_SIZE;
	stream->data_size += bytes;

	if (pos + bytes > BACKLOG_SIZE) {
		size_t to_copy = BACKLOG_SIZE - pos;
		memcpy(stream->data + pos, data, to_copy);
		data += to_copy;
		bytes -= to_copy;
		pos = 0;
	}

	memcpy(stream->data + pos, data, bytes);
	stream->wake_up_all_clients();
}

// See the .h file for postconditions after this function.	
void Server::process_client(Client *client)
{
	switch (client->state) {
	case Client::READING_REQUEST: {
read_request_again:
		// Try to read more of the request.
		char buf[1024];
		int ret;
		do {
			ret = read(client->sock, buf, sizeof(buf));
		} while (ret == -1 && errno == EINTR);

		if (ret == -1 && errno == EAGAIN) {
			// No more data right now. Nothing to do.
			// This is postcondition #2.
			return;
		}
		if (ret == -1) {
			perror("read");
			close_client(client);
			return;
		}
		if (ret == 0) {
			// OK, the socket is closed.
			close_client(client);
			return;
		}

		RequestParseStatus status = wait_for_double_newline(&client->request, buf, ret);
	
		switch (status) {
		case RP_OUT_OF_SPACE:
			fprintf(stderr, "WARNING: fd %d sent overlong request!\n", client->sock);
			close_client(client);
			return;
		case RP_NOT_FINISHED_YET:
			// OK, we don't have the entire header yet. Fine; we'll get it later.
			// See if there's more data for us.
			goto read_request_again;
		case RP_EXTRA_DATA:
			fprintf(stderr, "WARNING: fd %d had junk data after request!\n", client->sock);
			close_client(client);
			return;
		case RP_FINISHED:
			break;
		}

		assert(status == RP_FINISHED);

		int error_code = parse_request(client);
		if (error_code == 200) {
			construct_header(client);
		} else {
			construct_error(client, error_code);
		}

		// We've changed states, so fall through.
		assert(client->state == Client::SENDING_ERROR ||
		       client->state == Client::SENDING_HEADER);
	}
	case Client::SENDING_ERROR:
	case Client::SENDING_HEADER: {
sending_header_or_error_again:
		int ret;
		do {
			ret = write(client->sock,
		                    client->header_or_error.data() + client->header_or_error_bytes_sent,
				    client->header_or_error.size() - client->header_or_error_bytes_sent);
		} while (ret == -1 && errno == EINTR);

		if (ret == -1 && errno == EAGAIN) {
			// We're out of socket space, so now we're at the “low edge” of epoll's
			// edge triggering. epoll will tell us when there is more room, so for now,
			// just return.
			// This is postcondition #4.
			return;
		}

		if (ret == -1) {
			// Error! Postcondition #1.
			perror("write");
			close_client(client);
			return;
		}
		
		client->header_or_error_bytes_sent += ret;
		assert(client->header_or_error_bytes_sent <= client->header_or_error.size());

		if (client->header_or_error_bytes_sent < client->header_or_error.size()) {
			// We haven't sent all yet. Fine; go another round.
			goto sending_header_or_error_again;
		}

		// We're done sending the header or error! Clear it to release some memory.
		client->header_or_error.clear();

		if (client->state == Client::SENDING_ERROR) {
			// We're done sending the error, so now close.	
			// This is postcondition #1.
			close_client(client);
			return;
		}

		// Start sending from the end. In other words, we won't send any of the backlog,
		// but we'll start sending immediately as we get data.
		// This is postcondition #3.
		client->state = Client::SENDING_DATA;
		client->bytes_sent = client->stream->data_size;
		client->stream->put_client_to_sleep(client);
		return;
	}
	case Client::SENDING_DATA: {
		// See if there's some data we've lost. Ideally, we should drop to a block boundary,
		// but resync will be the mux's problem.
		Stream *stream = client->stream;
		size_t bytes_to_send = stream->data_size - client->bytes_sent;
		if (bytes_to_send == 0) {
			return;
		}
		if (bytes_to_send > BACKLOG_SIZE) {
			fprintf(stderr, "WARNING: fd %d lost %lld bytes, maybe too slow connection\n",
			        client->sock,
			        (long long int)(bytes_to_send - BACKLOG_SIZE));
			client->bytes_sent = stream->data_size - BACKLOG_SIZE;
			bytes_to_send = BACKLOG_SIZE;
		}

		// See if we need to split across the circular buffer.
		ssize_t ret;
		if ((client->bytes_sent % BACKLOG_SIZE) + bytes_to_send > BACKLOG_SIZE) {
			size_t bytes_first_part = BACKLOG_SIZE - (client->bytes_sent % BACKLOG_SIZE);

			iovec iov[2];
			iov[0].iov_base = const_cast<char *>(stream->data + (client->bytes_sent % BACKLOG_SIZE));
			iov[0].iov_len = bytes_first_part;

			iov[1].iov_base = const_cast<char *>(stream->data);
			iov[1].iov_len = bytes_to_send - bytes_first_part;

			do {
				ret = writev(client->sock, iov, 2);
			} while (ret == -1 && errno == EINTR);
		} else {
			do {
				ret = write(client->sock,
					    stream->data + (client->bytes_sent % BACKLOG_SIZE),
					    bytes_to_send);
			} while (ret == -1 && errno == EINTR);
		}
		if (ret == -1 && errno == EAGAIN) {
			// We're out of socket space, so return; epoll will wake us up
			// when there is more room.
			// This is postcondition #4.
			return;
		}
		if (ret == -1) {
			// Error, close; postcondition #1.
			perror("write/writev");
			close_client(client);
			return;
		}
		client->bytes_sent += ret;

		if (client->bytes_sent == stream->data_size) {
			// We don't have any more data for this client, so put it to sleep.
			// This is postcondition #3.
			stream->put_client_to_sleep(client);
		} else {
			// XXX: Do we need to go another round here to explicitly
			// get the EAGAIN?
		}
		break;
	}
	default:
		assert(false);
	}
}

int Server::parse_request(Client *client)
{
	vector<string> lines = split_lines(client->request);
	if (lines.empty()) {
		return 400;  // Bad request (empty).
	}

	vector<string> request_tokens = split_tokens(lines[0]);
	if (request_tokens.size() < 2) {
		return 400;  // Bad request (empty).
	}
	if (request_tokens[0] != "GET") {
		return 400;  // Should maybe be 405 instead?
	}
	if (streams.count(request_tokens[1]) == 0) {
		return 404;  // Not found.
	}

	client->stream_id = request_tokens[1];
	client->stream = find_stream(client->stream_id);
	if (client->stream->mark_pool != NULL) {
		client->fwmark = client->stream->mark_pool->get_mark();
	} else {
		client->fwmark = 0;  // No mark.
	}
	if (setsockopt(client->sock, SOL_SOCKET, SO_MARK, &client->fwmark, sizeof(client->fwmark)) == -1) {                          
		if (client->fwmark != 0) {
			perror("setsockopt(SO_MARK)");
		}
	}
	client->request.clear();

	return 200;  // OK!
}

void Server::construct_header(Client *client)
{
	client->header_or_error = find_stream(client->stream_id)->header;

	// Switch states.
	client->state = Client::SENDING_HEADER;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client->sock;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}
	
void Server::construct_error(Client *client, int error_code)
{
	char error[256];
	snprintf(error, 256, "HTTP/1.0 %d Error\r\nContent-type: text/plain\r\n\r\nSomething went wrong. Sorry.\r\n",
		error_code);
	client->header_or_error = error;

	// Switch states.
	client->state = Client::SENDING_ERROR;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client->sock;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}

template<class T>
void delete_from(vector<T> *v, T elem)
{
	typename vector<T>::iterator new_end = remove(v->begin(), v->end(), elem);
	v->erase(new_end, v->end());
}
	
void Server::close_client(Client *client)
{
	if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, client->sock, NULL) == -1) {
		perror("epoll_ctl(EPOLL_CTL_DEL)");
		exit(1);
	}

	// This client could be sleeping, so we'll need to fix that. (Argh, O(n).)
	if (client->stream != NULL) {
		delete_from(&client->stream->sleeping_clients, client);
		delete_from(&client->stream->to_process, client);
		if (client->stream->mark_pool != NULL) {
			int fwmark = client->fwmark;
			client->stream->mark_pool->release_mark(fwmark);
		}
	}

	// Bye-bye!
	int ret;
	do {
		ret = close(client->sock);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		perror("close");
	}

	clients.erase(client->sock);
}
	
Stream *Server::find_stream(const string &stream_id)
{
	map<string, Stream *>::iterator it = streams.find(stream_id);
	assert(it != streams.end());
	return it->second;
}

void Server::process_queued_data()
{
	MutexLock lock(&queued_data_mutex);

	for (size_t i = 0; i < queued_add_clients.size(); ++i) {
		add_client(queued_add_clients[i]);
	}
	queued_add_clients.clear();	
	
	for (map<string, string>::iterator queued_it = queued_data.begin();
	     queued_it != queued_data.end();
	     ++queued_it) {
		add_data(queued_it->first, queued_it->second.data(), queued_it->second.size());
	}
	queued_data.clear();
}
